"""
Module Name: lmsg.py
Author: Diogo Ferreira (ItsNotSoftware)
Date: May 8, 2024

Description:
    This module contains the python representation of YAML parsed messages.

License:
    Copyright (c) 2024 Diogo Ferreira. All rights reserved.
    This code is licensed under the MIT License.
"""

from pydantic import BaseModel, field_validator, model_validator
from lionsc.errors import *

_used_ids = []  # List to store the used IDs
_used_names = []  # List to store the used names


class MsgField(BaseModel):
    """
    Python representation of a message field in a LMsg

    Attributes:
        parent_msg_name (str): Name of the parent LMsg (used for error messages)
        name (str): Name of the field
        type (str): Type of the field
        size (int): Size of the field
        start (int): Start position of the field inside the payload array

    Methods:
        validate_size(cls, v: int) -> int: Validate the size of the field based on the type
        validate_type(self) -> MsgField: Validate the type of the field
    """

    parent_msg_name: str
    name: str
    type: str
    size: int
    start: int = 0
    buff_type: str = ""

    @field_validator("size")
    @classmethod
    def validate_size(cls, v: int) -> int:
        """
        Validate if the size of the field is valid

        Args:
            cls (MsgField): MsgField class
            v (int): size of the field

        Raises:
            ValueError: If the size is not between 0 and 248 bytes

        Returns:
            int: size of the field
        """

        if v < 0 or v > 248:
            raise ValueError("Size must be greater than 0 and less than 248 bytes")

        return v

    @model_validator(mode="after")
    def validate_type(self) -> "MsgField":
        """
        Validate the type specified by the user is valid

        Raises:
            InvalidTypeError: If the type is not valid

        Returns:
            MsgField: The current instance of the MsgField
        """

        if self.type not in [
            "bool",
            "int8_t",
            "uint8_t",
            "int16_t",
            "uint16_t",
            "int32_t",
            "uint32_t",
            "int64_t",
            "uint64_t",
            "float",
            "double",
            "string",
        ]:
            raise InvalidTypeError(self.parent_msg_name, self.name, self.type)

        return self


class LMsg(BaseModel):
    """
    Python representation of a LMsg (defined in YAML)

    Attributes:
        name (str): Name of the message
        id (int): ID of the message
        period (int): Period of the message (used to generate a constant for shceduling)
        fields (list[MsgField]): List of fields in the message

    Methods:
        validate_id(self): Validate the ID of the LMsg
        validate_name(cls, v: str) -> str: Validate the name of the LMsg
        validate_payload_size(cls, fields: list[MsgField]) -> list[MsgField]: Validate the payload size of the LMsg
        payload_size(self) -> int: Get the payload size of the LMsg
    """

    name: str
    id: int
    period: int
    fields: list[MsgField]

    @model_validator(mode="after")
    def validate_id(self) -> "LMsg":
        """
        Validate the ID of the LMsg

        Raises:
            DuplicateIdError: If the ID is already in use
            OutOfBoundsIdError: If the ID is out of bounds

        Returns:
            LMsg: The current instance of the LMsg
        """

        # Check if the ID is already in use
        if self.id in _used_ids:
            raise DuplicateIdError(
                self.name, self.id, [i for i in range(256) if i not in _used_ids][0]
            )

        # Check if the ID is out of bounds
        if self.id < 0 or self.id > 255:
            raise OutOfBoundsIdError(self.name, self.id)

        _used_ids.append(self.id)

        return self

    @model_validator(mode="after")
    def validate_payload_size(self) -> "LMsg":
        """
        Validate the payload size of the LMsg

        Raises:
            PayloadOverflowError: If the payload size is greater than 248 bytes

        Returns:
            LMsg: The current instance of the LMsg
        """

        payload_size = self.payload_size

        if payload_size > 248:
            raise PayloadOverflowError(self.name, payload_size)

        return self

    @field_validator("name")
    @classmethod
    def validate_name(cls, v: str) -> str:
        """
        Validate the name of the LMsg

        Note:
            ! This method does nothing because of the way the yaml module works. Duplicate names mean a previously defined message with the same name is overwritten.

        Args:
            cls (LMsg): LMsg class
            v (str): string containing the name

        Raises:
            DuplicateMsgNameError: If the name is already in use

        Returns:
            str: string containing the name
        """

        if v in _used_names:
            raise DuplicateMsgNameError(v)

        _used_names.append(v)

        return v

    @property
    def payload_size(self) -> int:
        """Get the payload size of the LMsg"""
        return sum(field.size for field in self.fields)



import pytest
from lionsc.code_generation.ts.ts_generator import TsGenerator
from lionsc.yaml_parser import YamlParser
from lionsc.lmsg import _used_ids, _used_names
import functools
import os


def clear_directory(directory_path):
    for filename in os.listdir(directory_path):
        file_path = os.path.join(directory_path, filename)

        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)


def reset(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _used_ids.clear()
        _used_names.clear()
        result = func(*args, **kwargs)
        return result

    return wrapper


def compare_files(file1, file2):
    with open(file1, "r") as f1, open(file2, "r") as f2:
        content1 = f1.read()
        content2 = f2.read()
    return content1 == content2


@reset
def test_ts_gen():
    msg_files_dir = "tests/test_files/single_lmsg_file2"
    output_dir = "tests/test_files/single_lmsg_file2/output"

    parser = YamlParser(msg_files_dir)
    generator = TsGenerator(output_dir)

    for filename, msgs in parser.parse_file():
        generator.generate_msg_files(filename, msgs)

    expected_files = [
        "lions.ts",
        "multiple_lmsg.ts",
    ]

    for file_name in expected_files:
        output_file = output_dir + "/" + file_name
        expected_file = msg_files_dir + "/expected_output/ts/" + file_name

        assert compare_files(
            output_file, expected_file
        ), f"{file_name} does not match expected output"

    clear_directory(output_dir)



import pytest
from lionsc.lmsg import LMsg, MsgField, _used_ids, _used_names
from lionsc.errors import *


def reset(func):
    def wrapper(*args, **kwargs):
        _used_ids.clear()
        _used_names.clear()
        result = func(*args, **kwargs)
        return result

    return wrapper


# Test valid LMsg creation
@reset
def test_valid_lmsg_creation():
    fields = [
        {
            "parent_msg_name": "msg1",
            "name": "field1",
            "type": "uint8_t",
            "size": 1,
            "start": 0,
        },
        {
            "parent_msg_name": "msg1",
            "name": "field2",
            "type": "uint16_t",
            "size": 2,
            "start": 1,
        },
    ]
    lmsg = LMsg(
        id=1, name="Test", period=100, fields=[MsgField(**field) for field in fields]
    )
    assert True


# Test duplicate LMsg ID
@reset
def test_duplicate_lmsg_id():
    LMsg(id=1, name="Test", period=100, fields=[])
    with pytest.raises(DuplicateIdError):
        LMsg(id=1, name="Test2", period=200, fields=[])


# Test duplicate msg name
@reset
def test_duplicate_lmsg_name():
    LMsg(id=1, name="Test", period=100, fields=[])
    with pytest.raises(DuplicateMsgNameError):
        LMsg(id=2, name="Test", period=200, fields=[])


# Test LMsg ID out of bounds
@reset
def test_lmsg_id_out_of_bounds():
    with pytest.raises(OutOfBoundsIdError):
        LMsg(id=256, name="Test", period=100, fields=[])


# Test valid field creation
@reset
def test_valid_field_creation():
    field = MsgField(parent_msg_name="msg", name="field1", type="uint8_t", size=1)
    assert field


# Test invalid field type
@reset
def test_invalid_field_type():
    with pytest.raises(ValueError):
        MsgField(name="field1", type="invalid_type", size=1)


# Test invalid field size for type
@reset
def test_invalid_field_size_for_type():
    with pytest.raises(ValueError):
        MsgField(name="field1", type="uint8_t", size=-2)


# Test invalid payload size
@reset
def test_payload_overflow():
    with pytest.raises(PayloadOverflowError):
        fields = [
            {
                "parent_msg_name": "msg1",
                "name": f"field{i}",
                "type": "uint8_t",
                "size": 100,
            }
            for i in range(10)
        ]  # Total payload size <= 248
        lmsg = LMsg(
            id=1,
            name="msg1",
            period=100,
            fields=[MsgField(**field) for field in fields],
        )


# Test valid payload size
@reset
def test_valid_payload_size():
    fields = [
        {"parent_msg_name": "msg1", "name": f"field{i}", "type": "uint8_t", "size": 1}
        for i in range(10)
    ]  # Total payload size <= 248
    lmsg = LMsg(
        id=1, name="Test", period=100, fields=[MsgField(**field) for field in fields]
    )
    assert lmsg



from jinja2 import Environment, FileSystemLoader
from lionsc.lmsg import LMsg
import time
from colorama import Fore, Style
import os


def print_generation_status(output_dir, file_name):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{timestamp} - Successfully generated: {output_dir}/{file_name}")


class CppGenerator:
    def __init__(self, output_dir: str):
        """
        Constructor for the CppGenerator class

        Args:
            output_dir (str): The output directory where the generated files will be saved
        """

        self.output_dir = output_dir

        template_dir = os.path.dirname(__file__) + "/templates"
        self.env = Environment(loader=FileSystemLoader(template_dir))

        # Load the templates
        self.lions_hpp_template = self.env.get_template("lions.hpp.jinja")
        self.lions_cpp_template = self.env.get_template("lions.cpp.jinja")
        self.msg_hpp_template = self.env.get_template("msg.hpp.jinja")
        self.msg_cpp_template = self.env.get_template("msg.cpp.jinja")

        self.generate_lions_hpp()
        self.generate_lions_cpp()

    def generate_lions_hpp(self):
        """Generate the lions.hpp file"""

        with open(f"{self.output_dir}/lions.hpp", "w") as f:
            f.write(self.lions_hpp_template.render())
            print_generation_status(self.output_dir, "lions.hpp")

    def generate_lions_cpp(self):
        """Generate the lions.cpp file"""

        with open(f"{self.output_dir}/lions.cpp", "w") as f:
            f.write(self.lions_cpp_template.render())
            print_generation_status(self.output_dir, "lions.cpp")

    def generate_msg_files(self, filename: str, msgs: list[LMsg]):
        """
        Generate the .cpp and .hpp message files for the given message file (yaml)

        Args:
            filename (str): The filename of the message files
            msgs (list[LMsg]): The list of messages to generate
        """

        file = f"{self.output_dir}/{filename}"

        # Dictionary to pass to the jinja templates
        jinja_dict = {"filename": filename, "msgs": msgs}

        for msg in msgs:
            for field in msg.fields:
                if field.type == "string":
                    field.type = "std::string"

        with open(f"{file}_lmsg.hpp", "w") as f_hpp, open(
            f"{file}_lmsg.cpp", "w"
        ) as f_cpp:
            # hpp
            f_hpp.write(self.msg_hpp_template.render(jinja_dict))
            print_generation_status(self.output_dir, f"{filename}_lmsg.hpp")

            # cpp
            f_cpp.write(self.msg_cpp_template.render(jinja_dict))
            print_generation_status(self.output_dir, f"{filename}_lmsg.cpp\n")



from jinja2 import Environment, FileSystemLoader
from lionsc.lmsg import LMsg
import time
from colorama import Fore, Style
import os


def print_generation_status(output_dir, file_name):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{timestamp} - Successfully generated: {output_dir}/{file_name}")


class CGenerator:
    def __init__(self, output_dir: str):
        """
        Constructor for the CGenerator class

        Args:
            output_dir (str): The output directory where the generated files will be saved
        """

        self.output_dir = output_dir

        template_dir = os.path.dirname(__file__) + "/templates"
        self.env = Environment(loader=FileSystemLoader(template_dir))

        # Load the templates
        self.lions_h_template = self.env.get_template("lions.h.jinja")
        self.lions_c_template = self.env.get_template("lions.c.jinja")
        self.msg_h_template = self.env.get_template("msg.h.jinja")
        self.msg_c_template = self.env.get_template("msg.c.jinja")

        self.generate_lions_h()
        self.generate_lions_c()

    def generate_lions_h(self):
        """Generate the lions.h file"""

        with open(f"{self.output_dir}/lions.h", "w") as f:
            f.write(self.lions_h_template.render())
            print_generation_status(self.output_dir, "lions.h")

    def generate_lions_c(self):
        """Generate the lions.c file"""

        with open(f"{self.output_dir}/lions.c", "w") as f:
            f.write(self.lions_c_template.render())
            print_generation_status(self.output_dir, "lions.c")

    def generate_msg_files(self, filename: str, msgs: list[LMsg]):
        """
        Generate the .c and .h message files for the given message file (yaml)

        Args:
            filename (str): The filename of the message files
            msgs (list[LMsg]): The list of messages to generate
        """

        file = f"{self.output_dir}/{filename}"

        # Dictionary to pass to the jinja templates
        jinja_dict = {"filename": filename, "msgs": msgs}

        for msg in msgs:
            for field in msg.fields:
                if field.type == "string":
                    field.type = "char*"

        with open(f"{file}_lmsg.h", "w") as f_h, open(f"{file}_lmsg.c", "w") as f_c:
            # h
            f_h.write(self.msg_h_template.render(jinja_dict))
            print_generation_status(self.output_dir, f"{filename}_lmsg.h")

            # c
            f_c.write(self.msg_c_template.render(jinja_dict))
            print_generation_status(self.output_dir, f"{filename}_lmsg.c\n")



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="lionsc",
    version="2.1.0",
    author="Diogo Ferreira",
    author_email="itsnotsoftware@gmail.com",
    description="LIONS is a communication protocol coupled with a compiler (lionsc), specifically designed for low-bandwidth IoT mesh and ad hoc networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={
        "lionsc": [
            "code_generation/cpp/templates/*.jinja",
            "code_generation/python/templates/*.jinja",
            "code_generation/js/templates/*.jinja",
            "code_generation/ts/templates/*.jinja",
        ]
    },
    url="https://github.com/ItsNotSoftware/lions",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=["pydantic", "pyyaml", "jinja2", "pytest", "colorama"],
    extras_require={
        "dev": ["twine>=4.0.2"],
    },
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "lionsc = lionsc.main:main",
        ],
    },
)
